/*
 * CCube.cpp
 *
 *  Created on: Dec 2, 2013
 *      Author: jono and miro
 */

#include "CCube.h"

/* Generic Constructor */
CCube::CCube() {

}

/* Copy constructor */
CCube::CCube( const CCube& cube ){
	/* Copies the location of the original cube */
	location_ = cube.location_;
}

/* Explicit Constructor */
CCube::CCube( std::vector< unsigned int > location ){
	location_ = location;
}

/* Destructor */
CCube::~CCube(){
}

std::vector< CCube > CCube::Subdivide(){
	/* Create a vector with four cubes corresponding to  a subdivision of the original cube */
	std::vector< CCube > subdivision(4, *this);
	/* Subdivide ( add one subdivision )*/
	for( unsigned int position = 0; position < 4; ++position )
		subdivision[ position ].location_.push_back( position );
	return subdivision;
}

/* returns the geometric position of the cube */
GeometricCube  CCube::GetGeometricRepresentation(){
	/*stores  the geometric representation of the cube */
	GeometricCube geom_cube;
        /* Sets the dimension of the geometric cube to 2 */
        geom_cube.dimension_=2;
	/* The minimum coordinates are set to the origin */
        geom_cube.coordinates_.push_back( Interval(0) );
        geom_cube.coordinates_.push_back( Interval(0) );

	/* The edge length of the unit cube is 1 */
	double edge_length = 1.0;
	for( unsigned int i = 0; i <  location_.size(); ++i ){
		/* In every subdivision the edge length decreases by half */
		edge_length = edge_length / 2.0;
		/* bit shifting operations to get the x and y position inside the given subdivision cube using the index */
		unsigned int x_coordinate = location_[ i ] & 1;
		unsigned int y_coordinate = ( (location_[ i ] & 2) >> 1 );
		/* if the cube is on the right side we add the edge length to the x coordinate */
                geom_cube.coordinates_[ 0 ] += ( (double)x_coordinate * edge_length );
		/* if the cube is on the top  we add the edge length to the y coordinate */
                geom_cube.coordinates_[ 1 ] += ( (double)y_coordinate * edge_length );
	}
	/* Update the maximum coordinates of the cube */
        for(int i = 0 ; i < 2 ; i++){
          geom_cube.coordinates_[i] += Interval(0,edge_length);
        }
//         geom_cube.coordinates_.push_back( ( geom_cube.min_coordinates_[ 0 ] + edge_length ) );
//         geom_cube.max_coordinates_.push_back( ( geom_cube.min_coordinates_[ 1 ] + edge_length ) );
	/* return the gemoetric position of the cube */
	return geom_cube;
}


void CCube::Print(){
	std::cout << "Location of the cube = [ ";
	for( unsigned int i = 0; i < location_.size(); ++ i )
		std::cout << location_[ i ] << " ";
	std::cout << "]\n";
}


/*
********************************
******   VERIFICATION    *******
******************************** 
*/

/*
This function verifies the cube using the input polynomial.
  -1 - the threshold intersects the function's image on a vertex ; This will cause the function Verified_Approximation to terminate.
   0 - cube was not verified
   1 - cube was verified
*/
int CCube::verify( CPolynomial poly , double threshold ){
  /* Create the geometric realization of the cube at hand */
  GeometricCube  geom_cube = GetGeometricRepresentation();
  /* We test whether the cube is ~~~ verified by sign~~~                */
  if( poly.value_test( geom_cube , threshold) )
    return 1;
  /* We test whether the cube is ~~~ verified by derivative~~~ */
  else{
    /* We compute and store the Jacobian of the function on the cube */
    std::vector<bool> Jacobian = poly.derivative_test( geom_cube );
    /* We test whether the cube is ~~~ strongly verified by derivative~~~ */
    if(strong_derivative_test(Jacobian)){
      /* We test whether the cube is ~~~ weakly verified by derivative~~~  only if one of the derivatives is bounded away from zero. */
      return 1;
    }
    else if(Jacobian[0] || Jacobian[1]){
      if( weak_derivative_test(poly, Jacobian, threshold , geom_cube))
        return 1;
    }
  }
  /* If we have reached this point, then the cube has not been verified */ 
  if( check_vertices(poly,threshold , geom_cube ) ){
    /* We test whether the cube fails the vertex condition                  */
    return -1;
  }
  else{
    /* Otherwise, the cube has simply failed to be verified  */
    return 0;
  }
}

/*
  ***************************
  ******   PRIVATE    *******
  *************************** 
*/
  
/*
This function checks whether the cube is ~~strongly~~ verified by derivative.

Input is a list of boolean values.
Returns true  if all values == 1;
Returns false otherwise.
---
Current implementation assumes the 2-dimensional case.
*/
bool CCube::strong_derivative_test(std::vector< bool > Jacobian){
  /* If both directional derivatives are bounded away from zero, then the cube is verified */
  if( Jacobian[0] && Jacobian[1] ) 
    return true;
  /* Otherwise the cube is not strongly verified by derivative */
  else
    return false;
}
  

/*
This function checks whether the cube is ~~weakly~~ verified by derivative.
---
Current implementation heavily assumes the 2-dimensional case.
*/
bool CCube::weak_derivative_test(CPolynomial poly , std::vector< bool > Jacobian , double threshold, const GeometricCube &geom_cube ){
  /* If both derivatives fail to be bounded away from zero, then return false. */
  if(!(Jacobian[0] || Jacobian[1] ))
    return false;
  else{
    /* Retrieve the edges we must try to verify by sign */
    std::vector< GeometricCube > faces = Get_Weak_Faces( Jacobian ,geom_cube);
    /* If both the edges are verified by sign, then return true */
    if ( poly.value_test(faces[0],threshold) &&  poly.value_test(faces[1],threshold) )
      return true;
    /* Otherwise, the cube is not weakly verified by derivative */
    else
      return false;
  }
}





/*
This function returns the verticies of a 2dimensional geometric cube 

The verticies are ordered as follows:: (0,0) ; (0,1) ; (1,0) ; (1,1) 
*/
std::vector< GeometricCube > CCube::get_vertices( const GeometricCube &geom_cube ){
  /* We obtain a list of our verticies */
  std::vector< GeometricCube > list_of_vertices ;
  for(int i = 0 ; i < 2 ;i++){
    for(int j = 0 ; j< 2 ; j++ ){
      GeometricCube new_vertex;
      
      if(i==0)    /* x has both min values */ 
        new_vertex.coordinates_.push_back(Interval(geom_cube.coordinates_[0].lower()));
      else        /* x has both max values */
        new_vertex.coordinates_.push_back(Interval(geom_cube.coordinates_[0].upper()));
      
      if(j==0)    /* y has both min values */
        new_vertex.coordinates_.push_back(Interval(geom_cube.coordinates_[1].lower()));
      else        /* y has both max values */
        new_vertex.coordinates_.push_back(Interval(geom_cube.coordinates_[1].upper()));
      
      /* We add the vertex to the ouput list */
      list_of_vertices.push_back(new_vertex);
    }
  }
  return list_of_vertices;
}


/*
This function checks to see if the image of any of the vertices is zero. 
Returns *true*  if the image of a vertex is zero
Returns *false* if the image of every vertex is nonzero   
*/
bool CCube::check_vertices(CPolynomial poly , double threshold , const GeometricCube &geom_cube ){
  /* We initialize our output.*/
  bool output = false;
  /* We obtain a list of our verticies */
  std::vector< GeometricCube > list_of_vertices = get_vertices(geom_cube);
  /* We Check the image of each of the verticies */
  for(int i = 0 ; i < list_of_vertices.size();i++){
    if( !(poly.value_test(list_of_vertices[i],threshold)) )
      output = true;
  }
  return output;
}


/*
This function outputs the appropriate faces for the weak derivative test. 
---
Current implementation heavily assumes the 2-dimensional case. 
*/
std::vector< GeometricCube > CCube::Get_Weak_Faces(std::vector< bool >  Jacobian , const GeometricCube &geom_cube ){
  /* We get the geometric representation of our cube*/
//   GeometricCube geom_cube = GetGeometricRepresentaion();
  /*We inialize our output vector of geometric cubes */
  std::vector< GeometricCube > output_faces;
  /* There are two cases to handle */
  /* The x-, but not the y- directional derivative is bounded away from zero */
  if( Jacobian[0] && !(Jacobian[1])){
    for(int i =0 ; i<2 ;i++){
      /* We initialize an edge */
      GeometricCube new_edge;
      
      /* The x- coordinates for the faces are different, and must be treated separately */
      /* Here we use the minimal x-edge */
      if(i == 0)
        new_edge.coordinates_.push_back(Interval(geom_cube.coordinates_[0].lower()));
      
      /* Here we use the maximal x-edge */
      else
        new_edge.coordinates_.push_back(Interval(geom_cube.coordinates_[0].upper()));
      
      /* The y- coordinate is an entire interval */
      new_edge.coordinates_.push_back(geom_cube.coordinates_[1]);
      /* We now add the edge to the faces we will output */
      output_faces.push_back(new_edge);
    } 
  }
  /* The y-, but not the x- directional derivative is bounded away from zero */
  else if( !(Jacobian[0]) && Jacobian[1] ){
    for(int i =0 ; i<2 ;i++){
      /* We initialize an edge */
      GeometricCube new_edge;
      
      /* The x- coordinate is an entire interval */
      new_edge.coordinates_.push_back(geom_cube.coordinates_[0]);
      
      /* The y- coordinates for the faces are different, and must be treated separately 
      Here we use the minimal y-edge */ 
      if(i == 0)
        new_edge.coordinates_.push_back(Interval(geom_cube.coordinates_[1].lower()));
      
      /* Here we use the maximal y-edge */
      else
        new_edge.coordinates_.push_back(Interval(geom_cube.coordinates_[1].upper()));
      /* We now add the edge to the faces we will output */
      output_faces.push_back(new_edge);
    } 
  }
  else{ 
    std::cout << "GET_WEAK FACES FUNCTION RECEIVED UNEXPECTED INPUT. \n";
  }
  return output_faces;
}

/* 
The verticies are ordered as follows:: (0,0) ; (0,1) ; (1,0) ; (1,1)  

A boolean value of <true> means the vertex's image is above the threshold
A boolean value of <false> means the vertex's image is below the threshold 

A cell of a verified cube is included in the approximation if and only if all of its vertices are in the approximation,
*/
std::vector< bool >  CCube::local_approximation( CPolynomial poly , double threshold ){
  /* Initize the output */
  std::vector< bool >  output;
  /* Obtain the geometric representation of the cube */ 
  GeometricCube geom_cube = GetGeometricRepresentation();
  /* Retrieve the verticies of the Cube */
  std::vector< GeometricCube > vertices = get_vertices( geom_cube );
  /* Determine which verticies are above and below the threshold */
  for(int i = 0 ; i < vertices.size();i++){
    /* We check to see if the vertex is above the threshold */
    if( poly.value(vertices[i]).lower() > threshold  )
      output.push_back(1);
    else
      output.push_back(0);
  }
  return output;
}
