/*
 * CComplex.cpp
 *
 *  Created on: Dec 6, 2013
 *      Author: Jono
 */

#include "CFiltration.h"


/*
***************************
******   PRIVATE    *******
*************************** 
*/


/*
This function is recursively defined.
If lambda and mu have the same splitting path length, then we return mu, 
Otherwise we need to subdivide mu along with its boundary.
  afterwards, we call Refine again, using the arguments lambda and the unique cell in subdivided-mu which contains lambda.
*/
SFace    CFiltration::Refine(SFace  lambda, SFace  mu ){  
  /* We check to see if lambda and mu are the same size of cell */
  if( lambda.path_.size() == mu.path_.size()){
    /* mu has support equal to lambda */
    return mu;
  }
  else{
    /* We need to subdivide the node associated with mu */
    (*mu.pt_node_).Filtered_Subdivide();
    
    /* We need to find which child number of mu contains lambda */
    int correct_child;
    /* We will need to manipulate the splitting path of mu */
    std::vector < unsigned int > local_path = mu.path_;
    /* We find a child cube from the node containing mu, for which that cube contains lambda */
    for(int i = 0; i < 4 ; i ++){
      /* We update the cell's splitting path */
      local_path.push_back(i);
      /* We define an SFace to be the i^th child, a 2-cell */
      SFace child_cube = {local_path , (*mu.pt_node_).get_child(i) , 0};
      /* If our current cube intersects lambda, then we break, knowing that i was the correct_child, notin that there might be multiple correct children */
      if( Intersect_Cube( child_cube , lambda) ){
        correct_child =i;
        break;
      }
      else{
        /* otherwise we return the local_path to be was mu's path is */
        local_path.pop_back();
      }
    }
    /* We need to define the child cube again */
    SFace child_cube = {local_path , (*mu.pt_node_).get_child(correct_child) , 0};
    /* We determine which face contains lambda */
    std::vector < int > correct_face = Intersecting_Faces( child_cube, lambda );
    /* We change the face_number of child_cube to be the correct face */
    child_cube.face_number_ = correct_face[0];
    /* We free some memory */
    correct_face.clear();
    
    
    /* We now impliment the recursive part of the function Refine, using child_cube*/
    return Refine(lambda,child_cube);
  }
}



/*
Given an SFace with complete information from the input complex, we add it to the filtration complex.
This may require either the input cell or its carrier in the filtration to be refined.

THIS FUNCTION HAS TO MAKE SURE THAT IF TWO CELLS INTERSECT, THEN THE SMALLER ONE HAS AN EARLIER BIRTH TIME
*/
void    CFiltration::Add_Cell_To_Complex( SFace lambda, int threshold){
  /* We need to check if there exists on unique cell \mu in the filtration complex which contains \lambda */
  /* Here unique referes to the geometric representation. there may very well be multiple computer representations*/
  
  /* The first order of business is to obtain a cover (inside the filtration) of the input cell */
  std::vector< SFace > cover = Obtain_Carrier( lambda );
  
  
  /* The function Obtain_Carrier will return redundant cells . We need to test several cases */
  bool unique_carrier = Unique_Carrier( lambda, cover );
  /* We determine if the cover of lamda has a cell with equal support */ 
  bool equal_support = false;
  for(int i = 0 ; i < cover.size(); i++){
    if(Equal_Support( lambda , cover[i] ))
      equal_support =true;
  }

  /* There are three cases for adding a cell to complex. 
  1) The filtration needs to be refined in order to add the cell.
  2) There is a cell in the filtration with support equal to that of the input cell.
  3) The cell needs be refined. */ 
  
  
  /* If there is a unique carrier which strictly contains lambda, then we are in Case 3). */
  if( unique_carrier && !(equal_support) ){    
    /* We call the function Refine with respect to lambda and the cell in the cover with the longest splitting path */
    if( cover.size() == 1){
      Refine( lambda, cover[0] );
    }
    else{
      /* Otherwise lambda is a 1-cell, containing 2 elements in its cover */
      if( cover[0].path_.size() >cover[1].path_.size() ){
        Refine( lambda, cover[0] );
      }
      else{
        Refine( lambda, cover[1] );
      }
    }
    /* We update the cover of lambda  */
    cover = Obtain_Carrier( lambda );
  }

  /* We update all the cells in the cover so that their birth time is no greater than the input-threshold */
  /* Nevertheless, these cells will be removed when we call Create_CW_Complex, so it should not affect anything. */
  for(int i = 0 ; i < cover.size(); i++){
    /* We prevent possibly altering large cells */
    bool containment = Containment( cover[i] , lambda );
    /* if we are  dealing with a cell contained in lambda, then we alter its birth time */
    if( containment  ){
      /* we check if the current cell ( cover[i] ) has later birth time */
      if( cover[i].BirthTime() > threshold ){
        /* if so, we set the cell's birth time to the more recent, lower time (which is threshold) */
        cover[i].DefineBirthTime(threshold );
      }
    }
  }
  
}



/* 
Given a cell lambda, and a complex not necessarily containing lambda, there exists a cover of lambda. 
This function tests whether there is a unique geometric cell in the complex which contains lambda.
 
 Does not use splitting path information
 
 TODO move this function to SFace 
 
*/
bool  CFiltration::Unique_Carrier( SFace lambda, std::vector < SFace >  cover ){
  /* The cover provided is sent by Obtain_Carrier, and will contain redundant cells. We need to test several cases */
  bool unique_carrier ;
  if( lambda.face_number_ == 0 ){
    /* lambda is a 2-cell,  */
    if( cover.size() == 1){
      /* If there is only one cell in the cover of lambda, then that cell contains lambda */
      unique_carrier = true;
    }
    else
      unique_carrier = false;
  }
  else if( lambda.face_number_ < 5 ){
    /* lambda is a 1-cell,  */
    if( cover.size() < 3){
      /* If there are only one or two cells in the cover of lambda, then lambda has a unique geometric cell containing it. */
      unique_carrier = true;
    }
    else
      unique_carrier = false;
  }
  else{
    /* lambda is a 0-cell, there is always a unique carrier  */
    unique_carrier = true;
  }
  return unique_carrier;
}




/*
Suppose we are given a cell lambda in complex A, which intersects a cell mu from complex B.
This function tests whether mu \subseteq lambda.
*/
bool  CFiltration::Containment( SFace  mu , SFace  lambda){
  bool containment;
  /* We break this down into cases by the dimension on lambda */
  if( lambda.face_number_ ==0  ){
    /* lambda is a 2-cell */
    /* if mu is a 0- or 1-cell, then  mu \subseteq lambda, but moreover, */
    /*  mu \subseteq lambda iff the depth of mu is greater than or equal to the depth of lambda */
    if( !(lambda.path_.size() > mu.path_.size() ) ) { 
      containment = true;
    }
    else{
      containment = false;
    }
  }
  else if ( lambda.face_number_  < 5){
    /* lambda is a 1-cell, */
    /* if mu is a 2-cell, mu is not contained in lambda */
    if( mu.face_number_ == 0){
      containment = false;
    }
    else if(  mu.face_number_ >4 ){
      /* if mu is a 0-cell, then it is contained in lambda */
      containment = true;
    }
    else{
      /* Otherwise, mu is a 1-cell */
      /*  mu \subseteq lambda iff the depth of mu is greater than or equal to the depth of lambda */
      if( !(lambda.path_.size() > mu.path_.size() )  ) { 
        containment = true;
      }
      else{
        containment = false;
      }
    }
  }
  else{
    /* Otherwise Lambda is a vertex */ 
    /* Equal support is true if and only if  mu has dimension zero */
    if(mu.face_number_ > 4 ){
      containment = true;
    }
    else{
      containment = false;
    }
  }
  /* We return the output */
  return containment;
}


/*
Suppose we are given a cell lambda from complex A and a cell mu from complex B such that they have a non-zero intersection. 
This function tests whether that mu has support equal to lambda.
*/
bool  CFiltration::Equal_Support( SFace lambda, SFace  mu ){
  /* We check to see if mu is contained in lambda */
  bool mu_in_lambda = Containment( mu , lambda );
  /* We check to see if lambda is contained in mu */
  bool lambda_in_mu = Containment( lambda , mu);
  /* if mu is contained in lambda, and lambda is contained in mu, then they have equal support */
  if( lambda_in_mu  && mu_in_lambda ){
    return true;
  }
  else{
    return false;
  }
}


/*
Given an node from the filtration complex, we add all of its cells to the Cell_List.
Additionally, we store inside that node, the order in which each cell appears in the list (w/ respect to dimension).
*/
void    CFiltration::Add_Cube_To_List( CNode* node  ){
  /* We create a blank splitting path vector for use in making SFaces */
  std::vector < unsigned int > blank;
  /* We resize the node's export order to equal 9 */
  (*node).ResizeExport();
  /* We iterate through all of the nine faces of the cube */
  for(int i = 0 ; i < 9 ; i++){
    /* If a cell is in the complex, then we create an SFace corresponding to that cell, (splitting path doesn't matter )*/
    SFace local_cell = { blank , node , i};
    if( local_cell.Check_Cell() ){
      /* We create a variable to store the Export Index of the cell */
      int export_index;
      /* We handel the case according to the dimension of the cell */
      if( i == 0 ){
        /* We are dealing with a 2-cell */
        export_index = Two_Cell_List.size();
        /* We add that SFace to the list of the appropriate dimension */
        Two_Cell_List.push_back( local_cell );
      }
      else if( i <5 ){
        /* We are dealing with a 1-cell */
        export_index = One_Cell_List.size();
        /* We add that SFace to the list of the appropriate dimension */
        One_Cell_List.push_back( local_cell );
      }
      else{
        /* We are dealing with a 0-cell */
        export_index = Zero_Cell_List.size();
        /* We add that SFace to the list of the appropriate dimension */
        Zero_Cell_List.push_back( local_cell );
      }
      /* We store on that node, the order in the list in which that cell appears */
      local_cell.DefineExportOrder(export_index ); 
    }
  }
}


std::string   CFiltration::Write_Cell( SFace cell ){
  /* We initialize a stream which will convert integers to strings */ 
  std::ostringstream convert;   // stream used for the conversion
  /* We first write the birthtime of the cell */
  convert << cell.BirthTime();
  /* We delimit by a space   */
  convert << " ";
  
  /* We now get the boundary of the cell inside of the filtration */
  std::vector < SFace >  boundary_positive = Boundary( cell, true  );
  std::vector < SFace >  boundary_negative = Boundary( cell, false );
  
  /* We write the the output the size of both the positive and negative boundary elements.
  Note that they need not be equal */
  convert << boundary_positive.size() + boundary_negative.size();
  
  /* We now write the positive boundary elements */
  for( int i = 0 ; i < boundary_positive.size() ; i ++ ){
    convert << " " << 1 ;
    convert << " " << boundary_positive[i].ExportOrder();
    /* We check to see if the boundary is born after the cell */
    if( boundary_positive[i].BirthTime() > cell.BirthTime() ){
      std::cout << "CRITICAL ERROR; BOUNDARY BORN AFTER CELL";
    }
  }
  /* We now write the negative boundary elements */
  for( int i = 0 ; i < boundary_negative.size() ; i ++ ){
    convert << " " << -1 ;
    convert << " " << boundary_negative[i].ExportOrder(); 
    /* We check to see if the boundary is born after the cell */
    if( boundary_negative[i].BirthTime() > cell.BirthTime() ){
      std::cout << "CRITICAL ERROR; BOUNDARY BORN AFTER CELL";
    }
  }


  /* We output what is inside of our stream */
  return convert.str();
}



void    CFiltration::Build_List(void){
	/* In order to build our list of cubes, we must iterate through all of the cubes in the complex of our filtration */
	/* Initialize the stack of CNode pointers we must add to the cell list */
	std::stack< CNode* > NodeStack;
	/* We add the pointer to root to the stack */
	NodeStack.push(&root);
	while (!(NodeStack.empty())){
		/* We remove the top CNode from the stack */
		CNode * current_node = NodeStack.top();
		NodeStack.pop();
		/* we check if the current node is a leaf */
		if ((*current_node).leaf_check()){
			/* If the current node is a leaf, then we add all of the cells from that node to the list */
			Add_Cube_To_List(current_node);
		}
		else{
			/*If current node is not a leaf, then we add all of its children to the stack */
			for (unsigned int i = 0; i < 4; i++){
				NodeStack.push((*current_node).get_child(i));
			}
		}
	}
}



/*
This function sets the birth time of lambda to be the minimal birth time of all the cells in the cover.
Functionally, lambda should be contained in each of the cells in this cover.
*/
void    CFiltration::Update_Minimal_Birth(SFace lambda, std::vector< SFace > cover){
	/* We calculate the minimum birth time of all the cells in the cover, which should include lambda */
	int minimum_birth_time = lambda.BirthTime();
	/* We iterate through the cells in the cover */
	for (int i = 0; i < cover.size(); i++){
		/* If the current cell has a lower birth time, then we update our minimum birth time */
		if (cover[i].BirthTime() < minimum_birth_time){
			minimum_birth_time = cover[i].BirthTime();
		}
	}
	/* We set the Birth time of lambda equal to the minimum birth time */
	lambda.DefineBirthTime(minimum_birth_time);
}



void  CFiltration::Update_Cell(SFace lambda){
	/* We break this down into cases by the dimension of lambda */
	if (lambda.face_number_ == 0){
		/* If lambda is a 2-cell, then we don't need to do anything */
	}
	else{
		/* We obtain a cover of lambda */
		std::vector < SFace > cover = Redundant_Cells(lambda);
		if (lambda.face_number_ < 5){
			/* Lambda is a 1-cell */
			/* We need to determine how big lambda is relative to the complex ; if lambda has a unique carrier */
			bool unique_carrier = Unique_Carrier(lambda, cover);
			/* If there is a unique carrier, then is small/normal sized, and should be updated */
			if (unique_carrier){
				Update_Minimal_Birth(lambda, cover);
			}
			/* If unique carrier is false, then we do nothing */
		}
		else{
			/* Lambda is a 0-cell */
			Update_Minimal_Birth(lambda, cover);
		}
	}

}




void    CFiltration::Update_Complex(void){
	/* In order to build our list of cubes, we must iterate through all of the cubes in the complex of our filtration */
	/* Initialize the stack of CNode pointers we must add to the cell list */
	std::stack< CNode* > NodeStack;
	/* We add the pointer to root to the stack */
	NodeStack.push(&root);
	/* We will be creating SFaces, so we have to make a blank vector */
	std::vector < unsigned int > blank;
	while (!(NodeStack.empty())){
		/* We remove the top CNode from the stack */
		CNode * current_node = NodeStack.top();
		NodeStack.pop();
		/* we check if the current node is a leaf */
		if ((*current_node).leaf_check()){
			/* If the current node is a leaf, then we add all of the cells from that node to the list */
			for (int i = 0; i<9; i++){
				SFace current_cell = { blank, current_node, i };
				Update_Cell(current_cell);
			}
		}
		else{
			/*If current node is not a leaf, then we add all of its children to the stack */
			for (unsigned int i = 0; i < 4; i++){
				NodeStack.push((*current_node).get_child(i));
			}
		}
	}
}



/*
***************************
******   PUBLIC     *******
*************************** 
*/



/*
We iterate through the imput complex, and add all of the cells to the filtration complex.
*/
void CFiltration::Add_Complex( CComplex & in_complex , int threshold){
  /* We need an empty location vector to create SFaces with */
  std::vector < unsigned int > blank;
  /* We get the root from the input complex */
  CNode * in_root = in_complex.get_node(blank);
  
  
/* In order to Add_Complex, we must iterate through every cube in the INPUT COMPLEX */
  /* Initialize the stack of CNode pointers we must add to the filtration. */
  std::stack< CNode* > NodeStack;
  /* We add the pointer to root to the stack */
  NodeStack.push(in_root);
  while( !(NodeStack.empty())){
    /* We remove the top CNode from the stack */
    CNode * current_node = NodeStack.top();
    NodeStack.pop();
    /* we check if the current node is a leaf */
    if( (*current_node).leaf_check()){
      /* If the current node is a leaf, then we determine its splitting path */
      std::vector < unsigned int > local_path = in_complex.get_location(current_node);
      /* We iterate through the different faces of the cube, and if they are in the input complex, then we add that cell to the filtration */
      for( int j = 0 ; j < 9 ; j++){ 
        /* If a given cell is in the complex, then we want to add it to the filtration */
        if( (*current_node).Check_Cell(j) ){
          /* We initialize the SFace we want to add to the complex */
          SFace local_cell = { local_path  , current_node , j};
          /* We add that local cell to the complex */
          Add_Cell_To_Complex(local_cell , threshold);
        }
      }
    }
    else{
      /*If current node is not a leaf, then we add all of its children to the stack */ 
      for( unsigned int i =0; i < 4 ;i++){
        NodeStack.push( (*current_node).get_child( i ) );
      }
    }
  }
  /* We now need to update the filtration complex, refining cells and updating their birth times. */
  Update_Complex();

}




void    CFiltration::Export( void ){
  /* First we must make the filtration a CW complex */
  Create_CW_Complex();
  /* We add all of the cells from the complex (tree) to the list */
  Build_List();
  /*We initialize an output stream */
  std::ofstream OutFile;
  /* The file we are writing to is called "FilteredComplex" */
  OutFile.open("FilteredComplex");

  /* We indicate the begining of a new dimension by writing a << -1 >> to the file. */
  OutFile << -1 ;
  /* We iterate through all of the zero-cells */
  for(int i = 0 ; i < Zero_Cell_List.size() ; i++){
    OutFile  << "\n" << Write_Cell( Zero_Cell_List[i]) ;
  }
  OutFile << "\n";
  
  /* We indicate the begining of a new dimension by writing a << -1 >> to the file. */
  OutFile << -1 ;
  /* We iterate through all of the one-cells */
  for(int i = 0 ; i < One_Cell_List.size() ; i++){
    OutFile  << "\n" << Write_Cell( One_Cell_List[i]) ;
  }
  OutFile << "\n";
  
  /* We indicate the begining of a new dimension by writing a << -1 >> to the file. */
  OutFile << -1 ;
  /* We iterate through all of the two-cells */
  for(int i = 0 ; i < Two_Cell_List.size() ; i++){
    OutFile  << "\n" << Write_Cell( Two_Cell_List[i]) ;
  }

  /* We close the file */
  OutFile.close();
}

/*
We define the initial complex. This needs to be run before one calls Add_Complex.
Adds the entire unit cube to the complex with birth time = max_steps+1. 
*/
void  CFiltration::Initialize( unsigned int max_steps){
  /* For << faces_ >> we create a boolean vector, of size 9, all entries are true */
  std::vector < bool > cell_signs( 9,true);
  /* We include all of the cells of the root into the complex */
  root.store_faces( cell_signs);
  
  /* We resize the birth_times vector so that it holds 9 elements */
  root.ResizeBirth();
  /* We set the birth times for all of the cells on the root equal to <<  max_steps +1 >> */
  for( int i = 0 ; i < 9 ; i ++ ){
    root.Set_Birth(i,max_steps +1);
  }
  
}


