/*
 * CPolynomial.cpp
 *
 *  Created on: Dec 2, 2013
 *      Author: jono and miro
 */

#include "CPolynomial.h"

/*
***************************
******   PRIVATE    *******
*************************** 
*/



std::vector < GeometricCube > CPolynomial::Subdivide( std::vector<GeometricCube> cube_list_in ){
  /* We initialize the output */
  std::vector < GeometricCube > output;
  
  for( int i = 0  ; i < cube_list_in.size() ; i++){
    /* For each cube in the first subdivision list, we subdivide that */
    std::vector < GeometricCube > local_cube_list = cube_list_in[i].Subdivide(); 
    for( int j = 0 ; j < local_cube_list.size() ; j++ ){
      /* We then add all the resulting cubes to the master cube list */
      output.push_back( local_cube_list[j] );
    }
  }
  return output;
}


std::vector < GeometricCube > CPolynomial::Multiple_Subdivisions(GeometricCube geom_input){
  /* We initialize a list of geometric cubes */
  std::vector < GeometricCube > cube_list;
  /* We add the input cube to the list */
  cube_list.push_back( geom_input);
  /* We subdivide the list depth-many times */
  for (int i = 0; i < Interval_Subdivisions_; i++){
    cube_list = Subdivide(cube_list);
  }
  
  return cube_list;
}

/* We subdivide the interval to get a closer estimate */
Interval CPolynomial::value_efficient( GeometricCube geom_input ){
  /* We subdivide the cube a certain number of times */
	std::vector < GeometricCube > cube_list = Multiple_Subdivisions(geom_input);
  

  /* We define a least upper bound, and a greatest lower bound */
  double upper_bound, lower_bound;
  for(int i = 0 ; i < cube_list.size(); i++){
    /* We iterate through each element in the list of cubes */
    /* We calculate and return the value of the function */
    Interval value_on_cube =  value( cube_list[i] );
    if( i == 0 ){
      /* We must begin with the first cube differently */
      upper_bound = value_on_cube.upper();
      lower_bound = value_on_cube.lower();
    }
    else{
      if( value_on_cube.lower() < lower_bound){
        /* We check to see if lower bound has decreased */
        lower_bound = value_on_cube.lower(); 
      }
      if( value_on_cube.upper() > upper_bound){
        /* We check to see if upper bound has increased */
        upper_bound = value_on_cube.upper(); 
      }
    }
  }
  Interval output = Interval( lower_bound , upper_bound );
  return output;
}

/* We subdivide the interval to get a closer estimate */
std::vector<Interval> CPolynomial::derivative_efficient( GeometricCube geom_input  ){
  /* We subdivide the cube a certain number of times */
	std::vector < GeometricCube > cube_list = Multiple_Subdivisions(geom_input);

  /* We define a least upper bound, and a greatest lower bound, for both x- and y- directional derivatives */
  std::vector<double> upper_bound, lower_bound;
  for(int i = 0 ; i < cube_list.size(); i++){
    /* We iterate through each element in the list of cubes */
    /* We calculate and return the value of the function */
    std::vector<Interval>  derivative_on_cube =  derivative( cube_list[i] );
    /* We iterate through each dimension, to improve our interval bounds */
    for(int k = 0 ; k < geom_input.dimension_ ; k++){
      if(i ==0){
        /* If we are at the first cube, then we set the upper and lower bounds accordingly */
        upper_bound.push_back( derivative_on_cube[k].upper());
        lower_bound.push_back( derivative_on_cube[k].lower());
      }
      else{
        if( derivative_on_cube[k].lower() < lower_bound[k]){
          /* We check to see if lower bound has decreased in dimension k */
          lower_bound[k] = derivative_on_cube[k].lower(); 
        }
        if( derivative_on_cube[k].upper() > upper_bound[k]){
          /* We check to see if upper bound has increased in dimension k */
          upper_bound[k] = derivative_on_cube[k].upper(); 
        }
      }
    }
  }

  /* Initialize the output vector */
  std::vector<Interval> output;
  /* We iterate through each dimension */
  for( int i =0; i< geom_input.dimension_ ; i++){
    /* We define an interval which has the improved directional derivative bounds along dimension i */
    Interval local_interval_output = Interval(  lower_bound[i], upper_bound[i] );
    /* We add this interval to our output vector (the Jacobian) */
    output.push_back(local_interval_output);
  }
  
  return output ;
}




/*
***************************
******   PUBLIC     *******
*************************** 
*/


/* 
Determines whether the GeometricCube is verified by sign, at the given threshold 
  0 -- Not Verified by sign
  1 -- Verified by sign
*/
bool CPolynomial::value_test( GeometricCube input, double threshold ){
  /* We determine what the image of the cube is. */
  Interval geom_value = value_efficient( input);
  /* Either the minimum of the Interval must be greater than the threshold, or its maximum must be less than the threshold */
  if( geom_value.lower() > threshold || geom_value.upper() < threshold)
    return 1;
  /* Otherwise the cube is not verified by sign */
  else
    return 0;
}

/* Determines whether the partial derivatives are bounded away from zero */
std::vector<bool> CPolynomial::derivative_test( GeometricCube input){
  /* Initalize the output */ 
  std::vector< bool> output;
  /* Calculate the Interval-Jacobian of the function */  
  std::vector<Interval> Jacobian = derivative_efficient(input);
  /* We iterate through each dimension */
  for( int i = 0 ; i< 2 ; i++){
    /* Either the minimum of the Interval must be greater than zero, or its maximum must be less than zero */
    if( Jacobian[i].lower() > 0 || Jacobian[i].upper() < 0)
      output.push_back(1);
    else
      output.push_back(0);
  }
  return output;
}

